﻿using System;
using System.Drawing;
using System.Windows.Forms;

using System.Reflection;
using System.CodeDom.Compiler;
using Microsoft.CSharp;

namespace howto_escape_time_polynomial
{
    public partial class Form1 : Form
    {
        public Form1()
        {
            InitializeComponent();
        }

        // Define the colors we will use.
        private Color[] FractalColors =
        {
            Color.Black, Color.Red, Color.Orange, Color.Yellow,
            Color.Green, Color.Cyan, Color.Blue, Color.Fuchsia,
        };
        int NumColors = 0;

        private enum SmoothingTypes
        {
            None,
            Smooth1,
            Smooth2,
        }
        private SmoothingTypes SmoothingType = SmoothingTypes.None;

        // Drawing parameters.
        private int MaxMagnitude;
        private int MaxIterations;

        // For smooth colors.
        private double LogEscape;

        // The bounds of interest for the different fractals.
        private RectangleF TestBounds = RectangleF.FromLTRB(-2.2f, -2.0f, 2.2f, 2.0f);

        // The largest interesting area for the current fractal.
        private RectangleF WorldMaxBounds;

        // The currently selected area.
        private RectangleF WorldBounds;

        // Set some initial values.
        private void Form1_Load(object sender, EventArgs e)
        {
            fractalPictureBox.Cursor = Cursors.Cross;

            // Set drawing parameters.
            NumColors = FractalColors.Length;
            MaxMagnitude = 2;
            MaxIterations = 64;
            LogEscape = Math.Log(MaxMagnitude);

            // Start with the Mandelbrot set.
            WorldMaxBounds = TestBounds;
            SetWorldBounds(TestBounds);

            fractalPictureBox.Refresh();
        }

        #region Coordinate systems

        // Set the world coordinate bounds as indicated,
        // adjust the aspect, and redraw the fractal.
        private void SetWorldBounds(RectangleF bounds)
        {
            // Set the world coordinate bounds.
            WorldBounds = bounds;
            AdjustBounds();
            DrawFractal();
        }

        // Adjust the world bounds so they have
        // the same aspect ratio as the PictureBox.
        private void AdjustBounds()
        {
            // Get the current world and device aspect ratios.
            float worldWidth = WorldBounds.Width;
            float worldHeight = WorldBounds.Height;
            float worldAspect = worldWidth / worldHeight;
            float deviceAspect =
                (float)fractalPictureBox.ClientSize.Width /
                (float)fractalPictureBox.ClientSize.Height;

            // Adjust the world sizes.
            if (worldAspect > deviceAspect)
            {
                // Too short and wide. Make taller.
                worldHeight = worldWidth / deviceAspect;
            }
            else
            {
                // Too tall and thin. Make wider.
                worldWidth = worldHeight * deviceAspect;
            }

            // Set the new bounds.
            float cx = WorldBounds.X + WorldBounds.Width / 2f;
            float cy = WorldBounds.Y + WorldBounds.Height / 2f;
            WorldBounds = new RectangleF(
                cx - worldWidth / 2f,
                cy - worldHeight / 2f,
                worldWidth,
                worldHeight);
        }

        // Convert from world to device coordinates.
        private PointF WtoD(PointF point)
        {
            float xscale = (point.X - WorldBounds.X) / WorldBounds.Width;
            float yscale = (point.Y - WorldBounds.Y) / WorldBounds.Height;

            float x = 0 + xscale * fractalPictureBox.ClientSize.Width;
            float y = 0 + yscale * fractalPictureBox.ClientSize.Height;
            return new PointF(x, y);
        }

        // Convert from device to world coordinates.
        private PointF DtoW(PointF point)
        {
            float xscale = (point.X - 0) / fractalPictureBox.ClientSize.Width;
            float yscale = (point.Y - 0) / fractalPictureBox.ClientSize.Height;
            float x = WorldBounds.X + xscale * WorldBounds.Width;
            float y = WorldBounds.Y + yscale * WorldBounds.Height;
            return new PointF(x, y);
        }

        #endregion Coordinate systems

        // Save the current image.
        private void mnuFileSaveAs_Click(object sender, EventArgs e)
        {
            if (fractalPictureBox.Image == null) return;
            if (sfdFractal.ShowDialog() == DialogResult.OK)
            {
                fractalPictureBox.Image.SaveImage(sfdFractal.FileName);
            }
        }

        private void mnuFileExit_Click(object sender, EventArgs e)
        {
            Close();
        }

        private void mnuSettingsNoSmoothing_Click(object sender, EventArgs e)
        {
            SelectSmoothing(sender, SmoothingTypes.None);
        }
        private void mnuSettingsSmooth1_Click(object sender, EventArgs e)
        {
            SelectSmoothing(sender, SmoothingTypes.Smooth1);
        }

        private void mnuSettingsSmooth2_Click(object sender, EventArgs e)
        {
            SelectSmoothing(sender, SmoothingTypes.Smooth2);
        }
        private void SelectSmoothing(object sender, SmoothingTypes smoothingType)
        {
            // Check the correct menu item.
            ToolStripMenuItem[] items =
            {
                mnuSettingsNoSmoothing,
                mnuSettingsSmooth1,
                mnuSettingsSmooth2,
            };
            SelectMenuItem(sender, items);
            SmoothingType = smoothingType;

            DrawFractal();
        }

        private void mnuScaleRedraw_Click(object sender, EventArgs e)
        {
            DrawFractal();
        }

        private void mnuScaleX2_Click(object sender, EventArgs e)
        {
            ScaleOut(2);
        }

        private void mnuScaleX4_Click(object sender, EventArgs e)
        {
            ScaleOut(4);
        }

        private void mnuScaleX8_Click(object sender, EventArgs e)
        {
            ScaleOut(8);
        }

        private void ScaleOut(float scale)
        {
            // Get the world coordinate center.
            float cx = WorldBounds.X + WorldBounds.Width / 2f;
            float cy = WorldBounds.Y + WorldBounds.Height / 2f;

            // Calculate the new width and height.
            float width = WorldBounds.Width * scale;
            float height = WorldBounds.Height * scale;

            // Set the new world bounds.
            WorldBounds = new RectangleF(
                cx - width / 2,
                cy - height / 2,
                width, height);

            // Redraw.
            DrawFractal();
        }

        private void mnuScaleFullScale_Click(object sender, EventArgs e)
        {
            WorldBounds = TestBounds;
            DrawFractal();
        }

        private void mnuScaleEnterSelectedArea_Click(object sender, EventArgs e)
        {
            InputForm dlg = new InputForm();
            dlg.Text = "Area to Select";
            dlg.captionLabel.Text = "xmin, ymin, xmax, ymax:";
            dlg.valueTextBox.Text =
                string.Format("{0}, {1}, {2}, {3}",
                    WorldBounds.Left, WorldBounds.Top,
                    WorldBounds.Right, WorldBounds.Bottom);

            if (dlg.ShowDialog() == DialogResult.OK)
            {
                string[] fields = dlg.valueTextBox.Text.Split(',');
                float xmin = float.Parse(fields[0]);
                float ymin = float.Parse(fields[1]);
                float xmax = float.Parse(fields[2]);
                float ymax = float.Parse(fields[3]);
                WorldBounds = new RectangleF(xmin, ymin, xmax - xmin, ymax - ymin);
                DrawFractal();
            }
        }

        // Check the selected menu item and uncheck the others.
        private void SelectMenuItem(object sender, ToolStripMenuItem[] items)
        {
            ToolStripMenuItem selected = sender as ToolStripMenuItem;
            foreach (ToolStripMenuItem item in items)
                item.Checked = (item == selected);
        }

        #region Mouse events

        private Point StartPoint, EndPoint;
        private bool SelectingArea = false;

        // Start selecting an area.
        private void fractalPictureBox_MouseDown(object sender, MouseEventArgs e)
        {
            StartPoint = e.Location;
            EndPoint = e.Location;

            fractalPictureBox.MouseMove += fractalPictureBox_MouseMove;
            fractalPictureBox.MouseUp += fractalPictureBox_MouseUp;

            SelectingArea = true;
            fractalPictureBox.Refresh();
        }


        // Continue selecting an area.
        private void fractalPictureBox_MouseMove(object sender, MouseEventArgs e)
        {
            EndPoint = e.Location;
            fractalPictureBox.Refresh();
        }

        // Finish selecting an area.
        private void fractalPictureBox_MouseUp(object sender, MouseEventArgs e)
        {
            SelectingArea = false;
            fractalPictureBox.MouseMove -= fractalPictureBox_MouseMove;
            fractalPictureBox.MouseUp -= fractalPictureBox_MouseUp;

            // Set the new world coordinate bounds.
            PointF wStartPoint = DtoW(StartPoint);
            PointF wEndPoint = DtoW(EndPoint);
            SetWorldBounds(wStartPoint.ToRectangle(wEndPoint));
        }

        // Draw the selection rectangle on top of the fractal.
        private void fractalPictureBox_Paint(object sender, PaintEventArgs e)
        {
            if (SelectingArea)
            {
                e.Graphics.DrawDashedBox(Color.Red, Color.Yellow,
                    2, 5, StartPoint, EndPoint);
            }
        }

        #endregion Mouse events

        #region Fractal drawing

        // Draw the selected fractal.
        private void DrawFractal()
        {
            // Hide any previous result.
            fractalPictureBox.Image = null;
            fractalPictureBox.Refresh();
            Cursor = Cursors.WaitCursor;

            // Make sure the aspect ratios match.
            AdjustBounds();
            Console.WriteLine(string.Format("{0}, {1}, {2}, {3}",
                WorldBounds.Left, WorldBounds.Top,
                WorldBounds.Right, WorldBounds.Bottom));

            // Find the X and Y scale factors.
            float dx = WorldBounds.Width /
                fractalPictureBox.ClientRectangle.Width;
            float dy = WorldBounds.Height /
                fractalPictureBox.ClientRectangle.Height;

            // Draw the fractal.
            // (x, y) will be the corresponding point in world coordinates.
            float xmin = WorldBounds.X;
            float ymin = WorldBounds.Y;

            // Get parameters.
            string z0 = txtZ0.Text;
            string equation = txtZn1.Text;

            // Make the bitmap.
            int width = fractalPictureBox.ClientSize.Width;
            int height = fractalPictureBox.ClientSize.Height;
            
            Bitmap bm = DrawFractal(xmin, ymin, dx, dy,
                width, height);

            // Display the result.
            fractalPictureBox.Image = bm;
            Cursor = Cursors.Default;
        }

        private Bitmap DrawFractal(float xmin, float ymin,
            float dx, float dy, int width, int height)
        {
            // Compile the evaluator function.
            MethodInfo method_info = GetMethodInfo(
                txtMaxSteps.Text, txtMaxMagnitude.Text,
                txtZ0.Text, txtZn1.Text);
            if (method_info == null) return null;

            // Make a Bitmap and associated Bitmap32 object.
            Bitmap bm = new Bitmap(width, height);
            Bitmap32 bm32 = new Bitmap32(bm);
            bm32.LockBitmap();

            // Loop over the pixels.
            object[] method_params = new object[2];
            float y = ymin;
            for (int iy = 0; iy < bm32.Height; iy++)
            {
                float x = xmin;
                for (int ix = 0; ix < bm32.Width; ix++)
                {
                    // Device coordinate point (ix, iy) corresponds
                    // to world coordinate point (x, y).

                    // Calculate the color for this point.
                    // Make the parameter list.
                    method_params[0] = x;
                    method_params[1] = y;

                    // Execute the method.
                    PointInfo point_info = (PointInfo)
                        method_info.Invoke(null, method_params);

                    // Set the pixel's color.
                    ColorPixel(bm32, ix, iy,
                        method_info, point_info.Z, point_info.StepNum);

                    // Move to the next point in world coordinates.
                    x += dx;
                }

                // Move to the next row in world coordinates.
                y += dy;
            }

            // Unlock the Bitmap32.
            bm32.UnlockBitmap();
            return bm;
        }

        private MethodInfo GetMethodInfo(
            string max_steps, string max_magnitude,
            string z0, string zn1)
        {
            // Compose the function.
            string function_text = @"
using howto_escape_time_polynomial;

public static class Evaluator
{
    public static PointInfo EvaluatePoint(float x, float y)
    {
        Complex i = new Complex(0, 1);
        Complex Zn = " + z0 + @";
        int step_num;
        for (step_num = 0; step_num < " + max_steps + @"; step_num++)
        {
            if (Zn.Magnitude > " + max_magnitude + @") break;

            // Calculate Zn+1.
            Complex ZnPlus1 = " + zn1 + @";
            Zn = ZnPlus1;
        }
        return new PointInfo(Zn, step_num);
    }
};";

            Console.WriteLine(function_text);

            // Generate a non-executable assembly in memory.
            CompilerParameters parameters = new CompilerParameters();
            parameters.GenerateInMemory = true;
            parameters.GenerateExecutable = false;
            parameters.ReferencedAssemblies.Add(
                Assembly.GetExecutingAssembly().Location);

            // Compile the code.
            CSharpCodeProvider code_provider =
                new CSharpCodeProvider();
            CompilerResults results =
                code_provider.CompileAssemblyFromSource(parameters, function_text);

            // If there are errors, display them.
            if (results.Errors.Count > 0)
            {
                string msg = "Error compiling the function.";
                foreach (CompilerError compiler_error in results.Errors)
                {
                    msg += "\n" + compiler_error.ErrorText;
                }
                MessageBox.Show(msg, "Expression Error",
                    MessageBoxButtons.OK, MessageBoxIcon.Error);
                return null;
            }
            else
            {
                // Get the Evaluator class type.
                Type evaluator_type = results.CompiledAssembly.GetType("Evaluator");
                //results.CompiledAssembly.Evidence.RemoveType(
                //    typeof(System.Security.Policy.Zone));
                //results.CompiledAssembly.Evidence.AddHost("internet");

                // Get a MethodInfo object describing the Evaluate method.
                MethodInfo method_info = evaluator_type.GetMethod("EvaluatePoint");

                // Return the MethodInfo.
                return method_info;
            }
        }

        // Set the pixel's color according to the
        // current smoothing model.
        private void ColorPixel(Bitmap32 bm32, int ix, int iy,
            MethodInfo method_info, Complex z, int step_num)
        {
            // If we exceeded MaxIterations, use black.
            if (step_num >= MaxIterations) step_num = 0;

            // Calculate the pixel's color.
            Color color;
            if (SmoothingType == SmoothingTypes.None)
            {
                color = FractalColors[step_num % NumColors];
            }
            else
            {
                color = SmoothColor(z, new Complex(), step_num);
            }

            // Set the pixel's color.
            bm32.SetPixel(ix, iy, color.R, color.G, color.B, 255);
        }

        // Return a smooth mandelbrot color.
        // See http://csharphelper.com/blog/2014/07/draw-a-mandelbrot-set-fractal-with-smoothly-shaded-colors-in-c/
        private Color SmoothColor(Complex z, Complex c, int step_num)
        {
            if (step_num == 0) return FractalColors[0];

            // Reduce the error in mu.
            for (int i = 0; i < 3; i++)
            {
                z = z * z + c;
                step_num++;
            }
            double mu = step_num + 1 - Math.Log(Math.Log(z.Magnitude)) / LogEscape;

            if (SmoothingType == SmoothingTypes.Smooth2)
                mu = mu / MaxIterations * NumColors;
            return MuToColor(mu);
        }

        private Color MuToColor(double mu)
        {
            int clr1 = (int)mu;
            double t2 = mu - clr1;
            double t1 = 1 - t2;
            clr1 = clr1 % NumColors;
            int clr2 = (clr1 + 1) % NumColors;

            double r = FractalColors[clr1].R * t1 + FractalColors[clr2].R * t2;
            double g = FractalColors[clr1].G * t1 + FractalColors[clr2].G * t2;
            double b = FractalColors[clr1].B * t1 + FractalColors[clr2].B * t2;
            return Color.FromArgb((byte)r, (byte)g, (byte)b);
        }

        #endregion Fractal drawing

        private void btnDraw_Click(object sender, EventArgs e)
        {
            DrawFractal();
        }

        private void mnuPresetMandelbrot_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "0";
            txtZn1.Text = "Zn.Pow(2) + x + y * i";
            DrawFractal();
        }

        private void mnuJuliaSetBasic_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) - i";
            DrawFractal();
        }

        private void mnuPresetJulia1_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (-0.4 + 0.6 * i)";
            DrawFractal();
        }

        private void mnuPresetJulia2_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (0.285 + 0 * i)";
            DrawFractal();
        }

        private void mnuPresetJulia3_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (0.285 + 0.01 * i)";
            DrawFractal();
        }

        private void mnuPresetJulia4_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (0.45 + 0.1428 * i)";
            DrawFractal();
        }

        private void mnuPresetJulia5_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (-0.70176 - 0.3842 * i)";
            DrawFractal();
        }

        private void mnuPresetJulia6_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (-0.835 - 0.2321 * i)";
            DrawFractal();
        }

        private void mnuPresetJulia7_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (-0.8 + 0.156 * i)";
            DrawFractal();
        }

        private void mnuPresetJulia8_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (-0.7269 + 0.1889 * i)";
            DrawFractal();
        }

        private void mnuPresetJulia9_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (0.0 - 0.8 * i)";
            DrawFractal();
        }

        private void mnuPresetJulia10_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (-0.74543 + 0.11301 * i)";
            DrawFractal();
        }

        private void mnuPresetJulia11_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (-0.75 + 0.11 * i)";
            DrawFractal();
        }

        private void mnuPresetJulia12_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "Zn.Pow(2) + (-0.1 + 0.651 * i)";
            DrawFractal();
        }

        private void mnuPresetJuliaish1_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "(1 - Zn.Pow(3) / 6) / (Zn - Zn.Pow(2) / 2).Pow(2) + 0.15";
            DrawFractal();
        }

        private void mnuPresetJuliaish2_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "1 - Zn.Pow(2) + Zn.Pow(5) / (2 + 4 * Zn)";
            DrawFractal();
        }

        private void mnuPresetJuliaish3_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "(1 - Zn.Pow(2) + Zn.Pow(5)) / (2 + 4 * Zn)";
            DrawFractal();
        }

        private void mnuPresetJuliaish4_Click(object sender, EventArgs e)
        {
            txtZ0.Text = "x + y * i";
            txtZn1.Text = "1 - Zn.Pow(2) + Zn.Pow(5) / (2 + 4 * Zn) + 0.2";
            DrawFractal();
        }
    }
}
